#include "cparser.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <setjmp.h>

#include <config.h>
#include <system.h>
#include <cpplib.h>
#include <obstack.h>

#include "keywords.c"
#include "gcc.c"

#define obstack_chunk_alloc xmalloc
#define obstack_chunk_free free

const char* progname;

static cpp_reader* reader;
struct line_maps line_table;
static cpp_dir* includes;

static struct obstack obst;

static uint32_t toknext;
static uint32_t tok, toklen;
static char* tokptr;

static char errmsg[256];
static jmp_buf errret;


static Sym syms[HASH_SIZE];
static Sym* cursym;
static Struct* curstr;

static void error(char* msg, ...) {
    char buf[256];
    va_list ap;
    va_start(ap, msg);
    vsprintf(buf, msg, ap);
    va_end(ap);

    sprintf(errmsg, "%s", buf);
    cpp_error(reader, CPP_DL_ERROR, errmsg);
//#define ABORT_IF_ERROR
#ifdef ABORT_IF_ERROR
    abort();
#else
    longjmp(errret, 1);
#endif
}

static void* mymalloc(int s) {
    return obstack_alloc(&obst, s);
}

static char* mstrcpy(const char* s) {
    char* r = (char*)xmalloc(strlen(s)+1);
    strcpy(r, s);
    return r;
}

static char* mstrcat(char* s, const char* a) {
    char* r;
    if (s) {
        r = (char*)xrealloc(s, strlen(s)+strlen(a)+1);
        strcat(r, a);
        return r;
    }
    else {
        return mstrcpy(a);
    }
}

static char* mstrcatfree(char* s, char* a) {
    char* r = mstrcat(s, a);
    free(a);
    return r;
}

static Type* type_alloc(Type t) {
    Type* type = (Type*)mymalloc(sizeof(Type));
    type->type = t.type;
    type->qual = t.qual;
    return type;
}

static const char* tok2str(int t) {
    static char buf[256];
    if (t < 128) {
        sprintf(buf, "%s", cpp_type2name(t));
        return buf;
    }
    else if (t == CPP_EOF) {
        return "EOF";
    }
    else if (t < TOKnum) {
        return keywords_str[t-129];
    }
    return "";
}

static char* type2str(const Type* type) {
    char* r = NULL;

    if (!type) return NULL;

    switch (type->qual & STORAGE_mask) {
    case STORAGE_static:
        r = mstrcpy("static ");
        break;
    case STORAGE_extern:
        r = mstrcpy("extern ");
        break;
    case STORAGE_inline:
        r = mstrcpy("inline ");
        break;
    case STORAGE_typedef:
        r = mstrcpy("typedef ");
        break;
    }

    if (type->qual & QUALIFY_const) {
        r = mstrcat(r, "const ");
    }
    if (type->qual & QUALIFY_volatile) {
        r = mstrcat(r, "volatile ");
    }
    if (type->qual & QUALIFY_restrict) {
        r = mstrcat(r, "restrict ");
    }

    if (chop_is_primitive(type)) {
        return mstrcat(r, tok2str((int)type->type));
    }
    else if (type->qual >> 8 == TYPE_typedef >> 8) {
        r = mstrcatfree(r, type2str((Type*)type->type));
        return r;
    }
    else if (type->qual >> 8 == TYPE_function >> 8) {
        Func* func = (Func*)type->type;
        Arg* a = func->args;
        r = mstrcat(r, "(");
        for (; a; ) {
            r = mstrcatfree(r, type2str(a->type));
            a = a->next;
            if (a) r = mstrcat(r, ",");
        }
        r = mstrcat(r, ")->");
        r = mstrcatfree(r, type2str(func->ret));
        return r;
    }
    else {
        switch (type->qual >> 8) {
        case 1: /* TYPE_pointer */
            r = mstrcat(r, "(");
            r = mstrcatfree(r, type2str((Type*)type->type));
            r = mstrcat(r, ")*");
            break;

        case 2: /* TYPE_array */
            r = mstrcat(r, "(");
            r = mstrcatfree(r, type2str((Type*)type->type));
            r = mstrcat(r, ")[]");
            break;

        case 4: /* TYPE_struct */
            r = mstrcat(r, "struct ");
            goto dump_struct_union;

        case 5: /* TYPE_union */
            r = mstrcat(r, "union ");

        dump_struct_union: {
                Struct* str = (Struct*)type->type;
                Member* mem;

                /* forward declaration */
                if (!str) {
                    r = mstrcat(r, "{ <unknown> }");
                    break;
                }

                r = mstrcat(r, "{\n");
                for (mem = str->mems; mem; mem = mem->next) {
                    char* p = type2str(mem->type);
                    r = mstrcat(r, p);
                    free(p);
                    r = mstrcat(r, " ");
                    r = mstrcat(r, mem->name);
                    r = mstrcat(r, "\n");
                }
                r = mstrcat(r, "}");
                break;
            }

        case 6: { /* TYPE_enum */
            Struct* str = (Struct*)type->type;
            Member* mem = str->mems;
            r = mstrcat(r, "enum ");
            r = mstrcat(r, str->name ? str->name : "(anon)");
            r = mstrcat(r, " {\n");
            for (; mem; mem = mem->next) {
                char* m = mem->name;
                r = mstrcat(r, m);
                r = mstrcat(r, mem->next ? ", " :"\n");
            }
            r = mstrcat(r, "}");

            break;
        }

        case 7: /* TYPE_ellipsis */
            r = mstrcpy("...");
            break;
        }
        return r;
    }
}

char* chop_type2str(const Type* type) {
    return type2str(type);
}

static Sym* lookup_sym() {
    int h = tok & (HASH_SIZE-1);
    if (tok < 128) {
        if (tok < N_TTYPES) {
            error("unexpected '%s'", cpp_type2name(tok));
        }
        else {
            error("unexpected stupid token(%d)", tok);
        }
    }
    while (syms[h].tok != 0) {
        if (syms[h].tok == tok && !strncmp(syms[h].name, tokptr, toklen)) {
            return &syms[h];
//            error("symbol redeclared (%s)", syms[h].name);
        }
        h++;
    }
    /* it's not good after parsing... */
    syms[h].tok = tok;
    syms[h].name = tokptr;
    return &syms[h];
}

static void pushtok(uint32_t t) {
    toknext = t;
}

static uint32_t calc_hash(char* p) {
    tok = 1;
    toklen = 0;
    tokptr = p;
    while (*p) {
        tok = tok * 131 + *p;
        p++;
        toklen++;
    }
    if (tok < 256) tok += 256;
    return tok;
}

static void skipto(uint32_t s1, uint32_t s2);
static void next() {
    const cpp_token* cpp_tok;

    if (toknext) {
        tok = toknext;
        toknext = 0;
        return;
    }

    while ((cpp_tok = cpp_get_token(reader))->type == CPP_PADDING);

    switch (cpp_tok->type) {
    case CPP_NAME: {
        calc_hash((char*)cpp_token_as_text(reader, cpp_tok));
        //printf("TOK: %s\n", p);
        tok = hash2tok(tok, tokptr);

        switch (tok) {
        case TOK__extension__:
            next();
            break;
        case TOK__const:
        case TOK__const__:
            tok = TOKconst;
            break;
/* attribute will be ignored using gcc.h
        case TOK__attribute:
        case TOK__attribute__: {
            next();
            if (tok != CPP_OPEN_PAREN) error("unknown __attribute__");
            next();
            if (tok != CPP_OPEN_PAREN) error("unknown __attribute__");
            skipto(CPP_CLOSE_PAREN, 0);
            skipto(CPP_CLOSE_PAREN, 0);
            next();
            break;
        }
*/
        case TOK__volatile:
        case TOK__volatile__:
            tok = TOKvolatile;
            break;
        case TOK__signed:
        case TOK__signed__:
            tok = TOKsigned;
            break;
        case TOK__inline:
        case TOK__inline__:
            tok = TOKinline;
            break;
        case TOK__restrict:
        case TOK__restrict__:
            tok = TOKrestrict;
            break;
        }

        break;
    }

    default: {
        tok = cpp_tok->type;
    }

    }
}

static void skipto(uint32_t s1, uint32_t s2) {
    while (1) {
        next();
        switch (tok) {
        case CPP_OPEN_BRACE: skipto(CPP_CLOSE_BRACE, 0); break;
        case CPP_OPEN_SQUARE: skipto(CPP_CLOSE_SQUARE, 0); break;
        case CPP_OPEN_PAREN: skipto(CPP_CLOSE_PAREN, 0); break;
        default: {
            if (s1 == tok || s2 == tok) return;
        }
        }
    }
}

static Type* declarator(Type type, int abstract);
static void declaration();

static void* usertype_specifiers(int is_enum) {
    Sym* s = NULL;
    char* name = NULL;

    next();

    if (tok > TOKnum) {
        s = lookup_sym();
        name = tokptr;
        next();
    }

    if (tok == CPP_OPEN_BRACE) {
        Struct* str = (Struct*)mymalloc(sizeof(Struct));
        Member** memp = &str->mems;
        str->name = name;

        if (is_enum) {
            next();
            while (1) {
                Member* mem = (Member*)mymalloc(sizeof(Member));
                mem->name = tokptr;
                mem->next = NULL;
                *memp = mem;
                memp = &mem->next;

                next();
                if (tok == CPP_EQ) {
                    skipto(CPP_COMMA, CPP_CLOSE_BRACE);
                }

                if (tok == CPP_COMMA) next();
                if (tok == CPP_CLOSE_BRACE) break;
            }
        }
        else {
            str->mems = NULL;
            next();
            while (tok != CPP_CLOSE_BRACE) {
                Struct* prev = curstr;
                curstr = str;
                pushtok(tok);
                declaration();
                curstr = prev;
                next();
            }

        }

        if (s) s->str = str;

        return str;
    }
    else {
        if (!s) error("expected ident or '{' after struct or union or enum");
        pushtok(tok);
        return s->str;
    }
}

static Type declaration_specifiers() {
    Type type = { NULL, 0 };
    int sign = 0;
    int user_type = 0;

next:
    next();
    switch (tok) {
    case TOKauto:
    case TOKregister:
        error("header should not contain auto or register");
        break;
    case TOKstatic:
        type.qual |= STORAGE_static;
        goto next;
    case TOKextern:
        type.qual |= STORAGE_extern;
        goto next;
    case TOKinline:
        type.qual |= STORAGE_inline;
        goto next;
    case TOKtypedef:
        type.qual |= STORAGE_typedef;
        goto next;
    case TOKconst:
        type.qual |= QUALIFY_const;
        goto next;
    case TOKvolatile:
        type.qual |= QUALIFY_volatile;
        goto next;
    case TOKrestrict:
        type.qual |= QUALIFY_restrict;
        goto next;

    case TOKint:
        if (!type.type) type.type = (void*)TOKint;
        goto next;

    case TOKvoid:
    case TOKchar:
    case TOKfloat:
    case TOKshort:
        type.type = (void*)tok;
        goto next;

    case TOKlong:
        if (type.type == (void*)TOKdouble) type.type = (void*)TOKlongdouble;
        else if (type.type == (void*)TOKlong) type.type = (void*)TOKlonglong;
        else type.type = (void*)TOKlong;
        goto next;

    case TOKdouble:
        if (type.type == (void*)TOKlong) type.type = (void*)TOKlongdouble;
        else type.type = (void*)TOKdouble;
        goto next;

    case TOKsigned:
        sign = 1;
        goto next;

    case TOKunsigned:
        sign = -1;
        goto next;

    case TOKstruct:
        type.qual |= TYPE_struct;
        type.type = usertype_specifiers(0);
        user_type = 1;
        goto next;

    case TOKunion:
        type.qual |= TYPE_union;
        type.type = usertype_specifiers(0);
        user_type = 1;
        goto next;

    case TOKenum:
        type.qual |= TYPE_enum;
        type.type = usertype_specifiers(1);
        user_type = 1;
        goto next;

    case CPP_ELLIPSIS:
        type.qual = TYPE_ellipsis;
        type.type = NULL;
        break;

    default: {
        if (user_type) {
            pushtok(tok);
            break;
        }
        if (type.type || sign) {
            switch ((int)type.type) {
            case TOKint:
                if (sign < 0) type.type = (void*)TOKuint;
                break;
            case TOKlong:
                if (sign < 0) type.type = (void*)TOKulong;
                break;
            case TOKlonglong:
                if (sign < 0) type.type = (void*)TOKulonglong;
                break;
            case TOKshort:
                if (sign < 0) type.type = (void*)TOKushort;
                break;
            case TOKchar:
                if (sign < 0) type.type = (void*)TOKuchar;
                else if (sign > 0) type.type = (void*)TOKschar;
                break;
            case 0:
                if (sign > 0) type.type = (void*)TOKint;
                else type.type = (void*)TOKuint;
                break;
            case TOKvoid:
            case TOKfloat:
            case TOKdouble:
            case TOKlongdouble:
                if (sign) {
                    error("cannot add signed or unsigned");
                }
                break;
            default:
                error("strange combination of primitive type");
            }
            pushtok(tok);
        }
        else {
            Sym* s = lookup_sym();
            if (s && s->sym) {
                Type* t = s->sym;
                if (!(t->qual & STORAGE_typedef))
                    error("%.*s: not type", toklen, tokptr);
                type.qual |= TYPE_typedef;
                type.type = t;
                user_type = 1;
                goto next;
            }
            else {
                type.type = (void*)TOKint;
                pushtok(tok);
            }
        }
    }
    }
    return type;
}

static Type* parameter_declaration() {
    Type t = declaration_specifiers();
    return declarator(t, 1);
}

static void parameter_list(Func* func) {
    func->args = NULL;
    next();
    if (tok == CPP_CLOSE_PAREN) {
        return;
    }
    else {
        Arg** ap = &func->args;
        pushtok(tok);
        while (1) {
            Type* t = parameter_declaration();

            if (t->type == (void*)TOKvoid) {
                next();
                if (tok == CPP_CLOSE_PAREN) break;
                error("unexpected token after void parameter.");
            }

            *ap = (Arg*)mymalloc(sizeof(Arg));
            (*ap)->next = NULL;
            (*ap)->type = t;
            ap = &(*ap)->next;

            next();

            if (tok == CPP_CLOSE_PAREN) break;
            if (tok != CPP_COMMA) error("unexpected token %s", tok2str(tok));
        }
    }
}

static void type_qualifier_list(Type* p) {
    next();
    while (tok == TOKconst || tok == TOKvolatile || tok == TOKrestrict) {
        switch (tok) {
        case TOKconst: p->qual |= QUALIFY_const; break;
        case TOKvolatile: p->qual |= QUALIFY_volatile; break;
        case TOKrestrict: p->qual |= QUALIFY_restrict; break;
        }
        next();
    }
}

static Type* declarator(Type type, int abstract) {
    Type* typep;
    Type** target;
    Type* ptr = NULL;

    int prev_storage_qual = type.qual & STORAGE_mask;
    type.qual &= ~STORAGE_mask;

    next();

    while (tok == CPP_MULT) {
        type.type = type_alloc(type);
        type.qual = TYPE_pointer;
        type_qualifier_list(&type);
    }

    typep = type_alloc(type);
    target = &typep;

    switch (tok) {
    case CPP_OPEN_PAREN: {
        Type* tmp;

        next();
        while (tok == CPP_MULT) {
            Type t = { ptr, TYPE_pointer | (type.qual & STORAGE_mask) };
            ptr = type_alloc(t);
            type_qualifier_list(ptr);
        }

        pushtok(tok);
        tmp = declarator(*typep, abstract);

        *target = typep = tmp;

        if (typep->qual == TYPE_function) {
            target = &((Func*)typep->type)->ret;
        }

        next();
        if (tok != CPP_CLOSE_PAREN) error("expected ')'");

        break;
    }
    case CPP_CLOSE_PAREN:
    case CPP_COMMA:
        if (!abstract) error("no identifier name");
        pushtok(tok);
        return *target;
    default: {
        if (abstract) break;

        if (curstr) {
            Member** mp;
            Member* mem = (Member*)mymalloc(sizeof(Member));
            mem->tok = tok;
            mem->name = tokptr;
            mem->type = typep;
            mem->bit = 0;
            mem->next = NULL;

            for (mp = &curstr->mems; *mp; mp = &(*mp)->next);
            *mp = mem;
        }
        else {
            Sym* sym = cursym = lookup_sym();
            sym->sym = typep;
            target = &sym->sym;
        }
    }
    }

    next();
    switch (tok) {
    case CPP_OPEN_SQUARE:
        (*target)->type = type_alloc(**target);
        (*target)->qual = TYPE_array;
        skipto(CPP_CLOSE_SQUARE, 0);
        break;
    case CPP_OPEN_PAREN: {
        Type ft;
        Func* func = (Func*)mymalloc(sizeof(Func));
        func->ret = *target;
        parameter_list(func);
        ft.type = func;
        ft.qual = TYPE_function;
        *target = type_alloc(ft);
        break;
    }
    default:
        pushtok(tok);
    }

    if (ptr) {
        Type* p = ptr;
        while (p->type) {
            p = (Type*)p->type;
        }
        p->type = *target;
        *target = ptr;
    }

    if (cursym) cursym->sym = *target;

    if (prev_storage_qual) (*target)->qual |= prev_storage_qual;
    return *target;
}

static void init_declarator(Type type) {
    cursym = NULL;
    declarator(type, 0);

    if (tok == CPP_EQ) {
        skipto(CPP_COMMA, CPP_SEMICOLON);
        pushtok(tok);
    }
}

static void declaration() {
    Type type = declaration_specifiers();

    next();

    /* implicit int type declaration*/
    if (tok == CPP_SEMICOLON || tok == CPP_COMMA) {
        if (type.qual == TYPE_struct ||
            type.qual == TYPE_union ||
            type.qual == TYPE_enum)
            return;

        error("not implemented");
        type.type = (void*)TOKint;
        type.qual = 0;
    }
    else {
        pushtok(tok);
    }

    while (1) {
        init_declarator(type);

        next();

        if (tok == CPP_COLON) {
            if (!curstr) error("floating colon");
            /* TODO: parse bitfield */
            skipto(CPP_SEMICOLON, 0);
        }

        if (tok == CPP_SEMICOLON) break;

        /* skip function definition */
        if (tok == CPP_OPEN_BRACE) {
            skipto(CPP_CLOSE_BRACE, 0);
            break;
        }

        if (tok != CPP_COMMA) error(", or ; expected");
    }
}

static void cparse_main() {
    while (1) {
        next();
        if (tok == CPP_EOF) {
            tokptr = NULL;
            return;
        }
        pushtok(tok);

        declaration();
    }
}

void chop_add_include(const char* p) {
    cpp_dir* d = XNEW(cpp_dir);
    d->next = includes;
    d->name = (char*)p;
    d->sysp = 1;
    d->construct = 0;
    d->user_supplied_p = 0;
    includes = d;
}

void chop_init() {
    obstack_init(&obst);

    linemap_init(&line_table);
    reader = cpp_create_reader(CLK_GNUC99, NULL, &line_table);

    chop_add_include("/usr/include");
    chop_add_include("/usr/local/include");
    chop_add_include(gcc_include_path);

    if (setjmp(errret)) {
        printf("Error when reading gcc.h: %s\n", errmsg);
        abort();
    }
    else {
        cpp_change_file(reader, LC_ENTER, "gcc.h");
        cpp_push_buffer(reader, gcch, gcch_size, 0);
        cparse_main();
    }
}

int chop_cparse(const char* filename) {
    progname = filename;

    cpp_set_include_chains(reader, includes, includes, false);

    if (setjmp(errret)) {
        //puts(errmsg);
        return 1;
    }
    else {
        if (!cpp_push_include(reader, filename)) return 1;
        cparse_main();
    }
    return 0;
}

Sym* chop_get_symbol(const char* name) {
    Sym* sym;
    calc_hash((char*)name);
    sym = lookup_sym();
    return sym;
}

char* chop_get_macro(const char* name) {
    cpp_hashnode* n = cpp_lookup(reader,
                                 (const unsigned char*)name, strlen(name));
    if (!n) return NULL;
    return index((char*)cpp_macro_definition(reader, n), ' ') + 1;
}

int chop_is_primitive(const Type* type) {
    return type->type && type->type < (void*)TOKnum;
}

void chop_iter_symbols(void (*callback) (const char* name, const Type* type)) {
    int i;
    for (i = 0; i < HASH_SIZE; i++) {
        if (!syms[i].sym) continue;
        callback(syms[i].name, syms[i].sym);
    }
}

typedef void(*macro_callback) (const char* name, const char* value);

static int iter_macros_callback(cpp_reader* r, cpp_hashnode* n, void* v) {
    macro_callback callback = (macro_callback)v;
    if (n->type == NT_MACRO && !(n->flags & NODE_BUILTIN)) {
        char* name = mstrcpy((char*)cpp_macro_definition(r, n));
        char* value = index(name, ' ');
        *value++ = '\0';
        callback(name, value);
        free(name);
    }
    return 1;
}

void chop_iter_macros(macro_callback callback) {
    cpp_forall_identifiers(reader, &iter_macros_callback, callback);
}

void chop_quit() {
    obstack_finish(&obst);
}
